#ifdef __RT__

#include <asm/io.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/rt_time.h>
#include <asm/rt_irq.h>
#include <linux/sched.h>
#include <linux/timex.h>

#define WAIT_LOCK 5

int rt_latch2;

static int only_linux, latency, oneshot_timer;

static struct rt_times rt_times, *active_rt_times;

static struct calibration_data tuned;

static void (*rt_linux_timer_hook)(void);


int rt_to_jiffie(void)
{
	int flags;
	int to_jiffie;
	
	r_save_flags(flags);
	r_cli();
	if (only_linux) {
		outb(0x00, 0x43);
		to_jiffie = inb(0x40);
		to_jiffie = - (to_jiffie | (inb(0x40)<<8));
		r_restore_flags(flags);
		return to_jiffie;
	}
	if (oneshot_timer) {
		to_jiffie =  active_rt_times->linux_time - rdtsc();
		to_jiffie = to_jiffie < 0 ? 0: - imuldiv(to_jiffie, FREQ_8254, tuned.cpu_freq);
	} else {
		to_jiffie = imuldiv(rdtsc() - active_rt_times->time_mark, FREQ_8254, tuned.cpu_freq);
		if ((to_jiffie += (int)(active_rt_times->tick_time - active_rt_times->linux_time)) > 0) {
			to_jiffie = 0;
		}
	}
	r_restore_flags(flags);
	return to_jiffie;
}


void rt_get_tuned_data(int cpu_freq, struct calibration_data *tuned_data)
{
#       define TWO61 ((double)0x1000000000000000LL)
#       define FREQ8254 ((double)FREQ_8254)
#       define SHIFT_FREQ8254 (((double)(1<<SHIFT_8254))*FREQ_8254)

	long linux_cr0, fpu_reg[27];
	int flags;

	if (only_linux) {
		only_linux = 0;
		tuned.cpu_freq = cpu_freq;
		tuned.latency = imuldiv(latency, cpu_freq, FREQ_8254) + tuned.setup_time_8254CPU;
		tuned.setup_time_8254 = imuldiv(tuned.setup_time_8254CPU, FREQ_8254, cpu_freq);

		r_save_flags(flags);
		r_cli();
		__asm__ __volatile__("movl %%cr0,%%eax": "=a" (linux_cr0): : "ax");
		__asm__ __volatile__("clts; fnsave %0" : "=m" (fpu_reg));
		tuned.oneshot_n2c = (TWO61*(double)cpu_freq)/1.0e9 + 0.49999999999999;
		tuned.oneshot_c2n = TWO61*1.0e9/(double)cpu_freq + 0.49999999999999;
		tuned.periodic_n2c = TWO61*SHIFT_FREQ8254/1.0e9 + 0.49999999999999;
		tuned.periodic_c2n = TWO61*1.0e9/SHIFT_FREQ8254 + 0.49999999999999;
		__asm__ __volatile__("frstor %0" : "=m" (fpu_reg));
		__asm__ __volatile__("movl %%eax,%%cr0": :"a" (linux_cr0):"ax");
		r_restore_flags(flags);
	}

	*tuned_data = tuned;
}


void rt_request_timer(void (*timer_handler)(void), int tick, struct rt_times *times) 
{
	int count;
	do {
		outb(0x00, 0x43);
		count = inb(0x40);
	} while	((count | (inb(0x40)<<8)) > WAIT_LOCK);
	if (tick > 0) {
		oneshot_timer = 0;
		times->tick_time = rt_times.tick_time;
		times->intr_time = rt_times.tick_time + tick;
		times->linux_tick = LATCH;
		times->linux_time = rt_times.linux_time;
		times->periodic_tick = tick;
		outb(0x34, 0x43);
	} else {
		oneshot_timer = 1;
		times->linux_tick = imuldiv(LATCH, tuned.cpu_freq, FREQ_8254);
		times->tick_time = rdtsc();
		times->intr_time = times->tick_time + times->linux_tick;
		times->linux_time = times->tick_time + times->linux_tick;
		times->periodic_tick = times->linux_tick;
		tick = LATCH;
		outb(0x30, 0x43);
	}
	outb(tick & 0xFF, 0x40);
	outb(tick >> 8, 0x40);
	active_rt_times = times;
	free_RTirq(0);
	request_RTirq(0, timer_handler);
	return;
}


static void linux_timer_handler(void)
{
	int count;

	rt_times.time_mark = rdtsc();
	SFREQ |= 1;
	rt_times.tick_time = rt_times.linux_time;
	rt_times.linux_time += LATCH;
	if (only_linux) {
		outb(0x00, 0x43);
		count = inb(0x40);
		latency = (3*latency + LATCH - (count | (inb(0x40)<<8)) + 1)>>2;
	}
	if (rt_linux_timer_hook) {
		(*rt_linux_timer_hook)();
	}
}


void rt_set_linux_timer_hook(void (*fun)(void))
{
	rt_linux_timer_hook = fun;
}


void rt_free_timer(void)
{
	outb(0x34, 0x43);
	outb(LATCH & 0xff, 0x40);
	outb(LATCH >> 8, 0x40);
	rt_times.tick_time = 0;
	rt_times.linux_time = LATCH;
	rt_times.periodic_tick = LATCH;
	active_rt_times = &rt_times;
	oneshot_timer = 0;
	free_RTirq(0);
	request_RTirq(0, linux_timer_handler);
}


void rt_time_init(void) 
{
#define LATCH2 50000
#define SHIFT 14

	RTIME t0_cpu;
	int flags, i;

	r_save_flags(flags);
	r_cli();

	rt_latch2 = LATCH2;
	outb_p(0xb6, 0x43);
	outb_p(rt_latch2 & 0xff, 0x42);
	outb(rt_latch2 >> 8, 0x42);
	outb((inb_p(0x61) & 0xfd) | 1, 0x61);
	outb(0x34, 0x43);

	t0_cpu = rdtsc();
	for (i = 1; i <= (1<<SHIFT); i++) {
		outb(LATCH & 0xff, 0x40);
		outb(LATCH >> 8, 0x40);
	}
	tuned.setup_time_8254CPU = ((int)(rdtsc() - t0_cpu) + (1<<(SHIFT-1)))>>SHIFT;

	only_linux = 1;
	rt_linux_timer_hook = 0;
	latency = 0;
	rt_free_timer();

	r_restore_flags(flags);
}

#endif /* __RT__ */
